/*
 * Copyright (C) 2000-2002 HighPoint Technologies, Inc.
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * File:	hptglb.h
 */
#ifndef _HPTGLB_H_
#define _HPTGLB_H_

#define MAX_MEMBERS       7    // Maximum members in an array
#define MAX_ADAPTERS      4    // Max Board we can supported in system
#define MAX_DEVICES       (MAX_ADAPTERS * 4)
#define MAX_V_DEVICE      MAX_DEVICES

#define TRUE             1
#define FALSE            0

typedef long LONG, *PLONG;
typedef unsigned long ULONG, *PULONG;
typedef unsigned long DWORD, *PDWORD, BOOL;
typedef unsigned int  UINT, *PUINT;
typedef unsigned char BYTE, *PBYTE;
typedef unsigned char UCHAR, *PUCHAR;
typedef UCHAR BOOLEAN;
typedef unsigned short USHORT, *PUSHORT;
typedef unsigned short WORD, *PWORD;
typedef void VOID, *PVOID;
typedef UCHAR * ADDRESS;

#define INLINE inline
#define IN
#define OUT
#define OPTIONAL

/*
 * Undef marcos which may conflict.
 */
#undef START_STOP

#pragma pack(1)

/*
 * IDE IO Port definition
 */
typedef struct _IDE_REGISTERS_1 {
	USHORT Data;               /* RW: Data port                  */
	UCHAR BlockCount;			 /* RW: Sector count               */
	UCHAR BlockNumber;			 /* RW: Sector number & LBA 0-7    */
	UCHAR CylinderLow;			 /* RW: Cylinder low & LBA 8-15    */
	UCHAR CylinderHigh;			 /* RW: Cylinder hign & LBA 16-23  */
	UCHAR DriveSelect;			 /* RW: Drive/head & LBA 24-27     */
	UCHAR Command;				 /* RO: Status WR:Command          */
} IDE_REGISTERS_1, *PIDE_REGISTERS_1;


/***************************************************************************
 *            IDE Control Register File 
 */

typedef struct _IDE_REGISTERS_2 {
	UCHAR AlternateStatus;     /* RW: device control port        */
} IDE_REGISTERS_2, *PIDE_REGISTERS_2;

//
// Identify data without the Reserved4.
//
typedef struct _IDENTIFY_DATA2 {
	USHORT GeneralConfiguration;            // 00 00
	USHORT NumberOfCylinders;               // 02  1
	USHORT Reserved1;                       // 04  2
	USHORT NumberOfHeads;                   // 06  3
	USHORT UnformattedBytesPerTrack;        // 08  4
	USHORT UnformattedBytesPerSector;       // 0A  5
	USHORT SectorsPerTrack;                 // 0C  6
	USHORT VendorUnique1[3];                // 0E  7-9
	USHORT SerialNumber[10];                // 14  10-19
	USHORT BufferType;                      // 28  20
	USHORT BufferSectorSize;                // 2A  21
	USHORT NumberOfEccBytes;                // 2C  22
	USHORT FirmwareRevision[4];             // 2E  23-26
	USHORT ModelNumber[20];                 // 36  27-46
	UCHAR  MaximumBlockTransfer;            // 5E  47
	UCHAR  VendorUnique2;                   // 5F
	USHORT DoubleWordIo;                    // 60  48
	USHORT Capabilities;                    // 62  49
	USHORT Reserved2;                       // 64  50
	UCHAR  VendorUnique3;                   // 66  51
	UCHAR  PioCycleTimingMode;              // 67
	UCHAR  VendorUnique4;                   // 68  52
	UCHAR  DmaCycleTimingMode;              // 69
	USHORT TranslationFieldsValid:1;        // 6A  53
	USHORT Reserved3:15;
	USHORT NumberOfCurrentCylinders;        // 6C  54
	USHORT NumberOfCurrentHeads;            // 6E  55
	USHORT CurrentSectorsPerTrack;          // 70  56
	ULONG  CurrentSectorCapacity;           // 72  57-58
	USHORT CurrentMultiSectorSetting;       //     59
	ULONG  UserAddressableSectors;          //     60-61
	USHORT SingleWordDMASupport : 8;        //     62
	USHORT SingleWordDMAActive : 8;
	USHORT MultiWordDMASupport : 8;         //     63
	USHORT MultiWordDMAActive : 8;
	USHORT AdvancedPIOModes : 8;            //     64
	USHORT Reserved4 : 8;
	USHORT MinimumMWXferCycleTime;          //     65
	USHORT RecommendedMWXferCycleTime;      //     66
	USHORT MinimumPIOCycleTime;             //     67
	USHORT MinimumPIOCycleTimeIORDY;        //     68
	USHORT Reserved5[2];                    //     69-70
	USHORT ReleaseTimeOverlapped;           //     71
	USHORT ReleaseTimeServiceCommand;       //     72
	USHORT MajorRevision;                   //     73
	USHORT MinorRevision;                   //     74
//    USHORT Reserved6[14];                   //     75-88
} IDENTIFY_DATA2, *PIDENTIFY_DATA2;

#define IDENTIFY_DATA_SIZE sizeof(IDENTIFY_DATA2)


typedef union _CDB {

	//
	// Standard 6-byte CDB
	//

	struct _CDB6READWRITE {
		UCHAR OperationCode;
		UCHAR LogicalBlockMsb1 : 5;
		UCHAR LogicalUnitNumber : 3;
		UCHAR LogicalBlockMsb0;
		UCHAR LogicalBlockLsb;
		UCHAR TransferBlocks;
		UCHAR Control;
	} CDB6READWRITE, *PCDB6READWRITE;

	//
	// Standard 10-byte CDB

	struct _CDB10 {
		UCHAR OperationCode;
		UCHAR Reserved1 : 5;
		UCHAR LogicalUnitNumber : 3;
		UCHAR LogicalBlockByte0;
		UCHAR LogicalBlockByte1;
		UCHAR LogicalBlockByte2;
		UCHAR LogicalBlockByte3;
		UCHAR Reserved2;
		UCHAR TransferBlocksMsb;
		UCHAR TransferBlocksLsb;
		UCHAR Control;
	} CDB10, *PCDB10;

	struct _START_STOP {
		UCHAR OperationCode;
		UCHAR Immediate: 1;
		UCHAR Reserved1 : 4;
		UCHAR LogicalUnitNumber : 3;
		UCHAR Reserved2[2];
		UCHAR Start : 1;
		UCHAR LoadEject : 1;
		UCHAR Reserved3 : 6;
		UCHAR Control;
	} START_STOP, *PSTART_STOP;

} CDB, *PCDB;

//
// SCSI CDB operation codes
//
#define SCSIOP_TEST_UNIT_READY     0x00
#define SCSIOP_REQUEST_SENSE       0x03
#define SCSIOP_READ6               0x08
#define SCSIOP_WRITE6              0x0A
#define SCSIOP_INQUIRY             0x12
#define SCSIOP_VERIFY6             0x13
#define SCSIOP_MODE_SENSE          0x1A
#define SCSIOP_START_STOP_UNIT     0x1B
#define SCSIOP_READ_CAPACITY       0x25
#define SCSIOP_READ                0x28
#define SCSIOP_WRITE               0x2A
#define SCSIOP_SEEK                0x2B
#define SCSIOP_VERIFY              0x2F


#define DRIVER_NAME				"Device Driver for HPT37x2 ATA RAID Controller"
#define COMPANY_NAME			"HighPoint Technologies, Inc."
#define CONTROLLER_NAME_HPT370	"HPT370 UDMA/ATA100 RAID Controller"
#define CONTROLLER_NAME_HPT372	"HPT372 UDMA/ATA133 RAID Controller"
#define CONTROLLER_NAME_HPT372A	"HPT372A UDMA/ATA133 RAID Controller"
#define PROC_DIR_NAME			"hpt37x2"
#define HPT_MAX_CMDS			124


typedef struct _PCI1_CFG_ADDR {
	unsigned    reg_num : 8;          // Register number
	unsigned    fun_num : 3;          // function number (0-1)
	unsigned    dev_num : 5;          // device number (0-20)
	unsigned    bus_num : 8;          // bus number (0)
	unsigned    reserved: 7;
	unsigned    enable  : 1;
}  PCI1_CFG_ADDR, *PPCI1_CFG_ADDR;

/*
 * PCI config Register offset
 */
#define REG_VID             0x00    // vender identification register
#define REG_DID             0x02    // device identification register
#define REG_PCICMD          0x04    // command register
#define REG_PCISTS          0x06    // PCI device status register
#define REG_RID             0x08    // revision identification register
#define REG_PI              0x09    // programming interface register
#define REG_SUBC            0x0a    // sub class code register
#define REG_BCC             0x0b    // base class code register
#define REG_MLT             0x0d    // master latency timer register
#define REG_HEDT            0x0e    // header type register
#define REG_IOPORT0         0x10    //
#define REG_IOPORT1         0x14    //
#define REG_BMIBA           0x20    // bus master interface base address reg
#define REG_IRQL			0x3C
#define REG_MISC            0x50    // brooklyn,MISC.control register

#define PCI_IOSEN           0x01    // Enable IO space
#define PCI_BMEN            0x04    // Enable IDE bus master


/***************************************************************************
 * Description:  Scatter-gather table
 */
typedef struct _SCAT_GATH
{
	ULONG  SgAddress;		 	
	USHORT SgSize;              
	USHORT SgFlag;              
}   SCAT_GATH, *PSCAT_GATH;

#define SG_FLAG_EOT         0x8000  
#define MAX_SG_DESCRIPTORS   17   //17 -- maximum 64K

/***************************************************************************
 * HPT Special
 */
#define HPT370_VENDORID 0x1103
#define HPT370_DEVICEID 0x0004
#define HPT372A_VENDORID 0x1103
#define HPT372A_DEVICEID 0x0005

/*
 * 370-370A timing 
 */
#define PIO_370MODE      0
#define DMA_370MODE      0x2
#define UDMA_370MODE     1

#define DATA_HIGH370(x)  ((ULONG)x)
#define DATA_LOW370(x)   ((ULONG)x << 4)
#define UDMA_CYCLE370(x) ((ULONG)x << 18)
#define DATA_PRE370(x)   ((ULONG)x << 22)
#define CTRL_ENA370(x)   ((ULONG)x << 28)

#define CMD_HIGH370(x)   ((ULONG)x << 9)
#define CMD_LOW370(x)    ((ULONG)x << 13)
#define CMD_PRE370(x)    ((ULONG)x << 25)

#define CLK50_PIO370     (CTRL_ENA370(PIO_370MODE)|CMD_PRE370(5)|CMD_LOW370(15)|CMD_HIGH370(10))

#define CLK50_370PIO0    (CLK50_PIO370|DATA_PRE370(3)|DATA_LOW370(8)|DATA_HIGH370(10))
#define CLK50_370PIO1    (CLK50_PIO370|DATA_PRE370(3)|DATA_LOW370(6)|DATA_HIGH370(5))
#define CLK50_370PIO2    (CLK50_PIO370|DATA_PRE370(2)|DATA_LOW370(5)|DATA_HIGH370(4))
#define CLK50_370PIO3    (CLK50_PIO370|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(3))
#define CLK50_370PIO4    (CLK50_PIO370|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))

#define CLK50_DMA370     (CTRL_ENA370(DMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(4)|CMD_HIGH370(1))
#define CLK50_370DMA0    (CLK50_DMA370|DATA_PRE370(2)|DATA_LOW370(14)|DATA_HIGH370(10))
#define CLK50_370DMA1    (CLK50_DMA370|DATA_PRE370(2)|DATA_LOW370(5)|DATA_HIGH370(4))
#define CLK50_370DMA2    (CLK50_DMA370|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))

#define CLK50_UDMA370    (CTRL_ENA370(UDMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(4)|CMD_HIGH370(1))
#define CLK50_370UDMA0   (CLK50_UDMA370|UDMA_CYCLE370(6)|DATA_LOW370(14)|DATA_HIGH370(10))
#define CLK50_370UDMA1   (CLK50_UDMA370|UDMA_CYCLE370(5)|DATA_LOW370(5)|DATA_HIGH370(4))
#define CLK50_370UDMA2   (CLK50_UDMA370|UDMA_CYCLE370(3)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK50_370UDMA3   (CLK50_UDMA370|UDMA_CYCLE370(3)|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK50_370UDMA4   (CLK50_UDMA370|UDMA_CYCLE370(11)|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK50_370UDMA5   (CLK50_UDMA370|UDMA_CYCLE370(1)|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))

#define CLK33_PIOCMD370  (CTRL_ENA370(PIO_370MODE)|CMD_PRE370(3)|CMD_LOW370(10)|CMD_HIGH370(7))
#define CLK33_370PIO0    (CLK33_PIOCMD370|DATA_PRE370(2)|DATA_LOW370(10)|DATA_HIGH370(7))
#define CLK33_370PIO1    (CLK33_PIOCMD370|DATA_PRE370(2)|DATA_LOW370(9)|DATA_HIGH370(3))
#define CLK33_370PIO2    (CLK33_PIOCMD370|DATA_PRE370(1)|DATA_LOW370(5)|DATA_HIGH370(3))
#define CLK33_370PIO3    (CLK33_PIOCMD370|DATA_PRE370(1)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK33_370PIO4    (CLK33_PIOCMD370|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))

#define CLK33_DMACMD370  (CTRL_ENA370(DMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(3)|CMD_HIGH370(6))
#define CLK33_370DMA0    (CLK33_DMACMD370|DATA_PRE370(1)|DATA_LOW370(9)|DATA_HIGH370(7))
#define CLK33_370DMA1    (CLK33_DMACMD370|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(3))
#define CLK33_370DMA2    (CLK33_DMACMD370|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))

#define CLK33_UDMACMD370 (CTRL_ENA370(UDMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(3)|CMD_HIGH370(1))
#define CLK33_370UDMA0   (CLK33_UDMACMD370|UDMA_CYCLE370(4)|DATA_PRE370(1)|DATA_LOW370(9)|DATA_HIGH370(7))
#define CLK33_370UDMA1   (CLK33_UDMACMD370|UDMA_CYCLE370(3)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(3))
#define CLK33_370UDMA2   (CLK33_UDMACMD370|UDMA_CYCLE370(2)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))
#define CLK33_370UDMA3   (CLK33_UDMACMD370|UDMA_CYCLE370(11)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))
#define CLK33_370UDMA4   (CLK33_UDMACMD370|UDMA_CYCLE370(1)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))
#define CLK33_370UDMA5   (CLK33_UDMACMD370|UDMA_CYCLE370(1)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))


/*
 * io commands
 */
#define IO_COMMAND_READ		1
#define IO_COMMAND_WRITE	2

/************************************************************************
 * shared data structures
 ************************************************************************/
 
/*
 * Chip Type
 */ 
#define CHIP_TYPE_HPT366      1
#define CHIP_TYPE_HPT368      2
#define CHIP_TYPE_HPT370      3
#define CHIP_TYPE_HPT370A     4
#define CHIP_TYPE_HPT370B     5
#define CHIP_TYPE_HPT374      6
#define CHIP_TYPE_HPT372      7
#define CHIP_TYPE_HPT372A     8

/*
 * ioctl structure
 */
#define HPT_IOCTL_MAGIC 0x1A2B3C4D

typedef struct _HPT_IOCTL_PARAM {
	DWORD Magic;				
	DWORD dwIoControlCode;       
	PVOID lpInBuffer;            
	DWORD nInBufferSize;      
	PVOID lpOutBuffer;          
	DWORD nOutBufferSize;    
	PDWORD lpBytesReturned; 
}
HPT_IOCTL_PARAM, *PHPT_IOCTL_PARAM;  	


/* ioctl command
 */
#define HPT_IOCTL_CMD	0xFE

//
// SRB Functions
//
#define SRB_FUNCTION_EXECUTE_SCSI           0x00
#define SRB_FUNCTION_IO_CONTROL             0x02
#define SRB_FUNCTION_SHUTDOWN               0x07
#define SRB_FUNCTION_FLUSH                  0x08
//
// SRB Status
//
#define SRB_STATUS_PENDING                  0x00
#define SRB_STATUS_SUCCESS                  0x01
#define SRB_STATUS_ABORTED                  0x02
#define SRB_STATUS_ERROR                    0x04
#define SRB_STATUS_BUSY                     0x05
#define SRB_STATUS_INVALID_REQUEST          0x06
#define SRB_STATUS_INVALID_PATH_ID          0x07
#define SRB_STATUS_NO_DEVICE                0x08
#define SRB_STATUS_SELECTION_TIMEOUT        0x0A
#define SRB_STATUS_BUS_RESET                0x0E
#define SRB_STATUS_NO_HBA                   0x11
#define SRB_STATUS_INVALID_LUN              0x20
#define SRB_STATUS_INVALID_TARGET_ID        0x21
#define SRB_STATUS_NEED_REQUEUE             0x24
//
// SRB Status Masks
//
#define SRB_STATUS_QUEUE_FROZEN             0x40
#define SRB_STATUS_AUTOSENSE_VALID          0x80

#define SRB_STATUS(Status) (Status & ~(SRB_STATUS_AUTOSENSE_VALID | SRB_STATUS_QUEUE_FROZEN))

//
// SRB Flag Bits
//
#define SRB_FLAGS_DATA_IN                   0x00000040
#define SRB_FLAGS_DATA_OUT                  0x00000080

typedef struct _SrbExtension {
	ULONG      StartLBA;   

    UCHAR      FirstMember;    
    UCHAR      LastMember;     
    UCHAR      InSameLine;     
    UCHAR      reserve1;

    USHORT     FirstSectors;   
    USHORT     FirstOffset;    

    USHORT     LastSectors;    
    USHORT     AllMemberBlocks;

	SCAT_GATH  ArraySg[MAX_SG_DESCRIPTORS]; 

	ADDRESS    DataBuffer;    
	USHORT     DataTransferLength;
	USHORT     SgFlags;    

	USHORT     JoinMembers;    
	USHORT     WaitInterrupt;  

	ULONG      Lba;            
	UCHAR      nSector;        
	UCHAR      SrbFlags;       

	struct _Device *pMaster;   

	ULONG       	WorkingFlags;

	Scsi_Cmnd       *pREQ; 				
	struct _Device  *pDev;				

} SrbExtension, *PSrbExtension;


/* SRB Working flags define area */
#define	SRB_WFLAGS_USE_INTERNAL_BUFFER		0x00000001 
#define	SRB_WFLAGS_IGNORE_ARRAY				0x00000002 
#define	SRB_WFLAGS_HAS_CALL_BACK			0x00000004 
#define	SRB_WFLAGS_MUST_DONE				0x00000008 
#define	SRB_WFLAGS_ON_MIRROR_DISK			0x00000010 
#define	SRB_WFLAGS_ON_SOURCE_DISK			0x00000020 
#define SRB_WFLAGS_ARRAY_IO_STARTED         0x10000000 
#define SRB_WFLAGS_WATCHTIMER_CALLED		0x20000000 
#define SRB_WFLAGS_USE_SG					0x40000000 

//
// SCSI I/O Request Block
//
typedef struct _SCSI_REQUEST_BLOCK {
	USHORT Length;
	UCHAR Function;
	UCHAR SrbStatus;
	UCHAR ScsiStatus;
	UCHAR CdbLength;
	UCHAR SenseInfoBufferLength;
	UCHAR reseved;
	ULONG SrbFlags;
	ULONG DataTransferLength;
	ULONG TimeOutValue;
	PVOID DataBuffer;
	PVOID SenseInfoBuffer;
	PVOID NextSrb;
	UCHAR Cdb[16];
	/* buffer to hold SrbExtension
	 * it must use 16-byte align.
	 */
	PVOID SrbExtension;
	UCHAR _SrbExtensionBuffer[sizeof(SrbExtension) + 16];
} SCSI_REQUEST_BLOCK, *PSCSI_REQUEST_BLOCK;

#define SCSI_REQUEST_BLOCK_SIZE sizeof(SCSI_REQUEST_BLOCK)

struct _Channel;
/***************************************************************************
 * Description: _Device
 */
typedef struct _Device {
	UINT    DeviceFlags;       
	UINT	DeviceFlags2;	   
	UCHAR   UnitId;            
	UCHAR   DeviceModeSetting; 
					
	UCHAR   ReadCmd;           
	UCHAR   WriteCmd;          
	UINT    MultiBlockSize;    

	USHORT  RealHeadXsect;     
	UCHAR   RealSector;        
	UCHAR   RealHeader;        

	UCHAR   bestPIO;           
	UCHAR   bestDMA;           
	UCHAR   bestUDMA;          
	UCHAR   Usable_Mode;       

	ULONG   HidenLBA;          
	struct _Channel *pChannel; 

	ULONG   capacity;          

	IDENTIFY_DATA2 IdentifyData;

	UCHAR   SmartCommand;      
	UCHAR   ReturningMediaStatus; 

	UCHAR   ArrayNum;		
	UCHAR   reserved1;
	USHORT  ArrayMask;         
	USHORT  reserved2;
	struct _VirtualDevice *pArray;

	struct{
		ULONG	nLastError;				
		struct _Device	*pNextErrorDevice;		
		UCHAR	Cdb[16];
	} stErrorLog;

	ULONG   RebuiltSector;

	UINT   IoCount; 
	UINT   ResetCount; 
	UINT   IoSuccess;	

} Device, *PDevice;

/*
 *  First device flags declare
 */
#define DFLAGS_LBA                  0x1    
#define DFLAGS_DMA                  0x2    
#define DFLAGS_MULTIPLE             0x4    
#define DFLAGS_ULTRA                0x8    

#define DFLAGS_DMAING               0x10   
#define DFLAGS_REQUEST_DMA          0x20   
#define DFLAGS_HIDEN_DISK           0x40   
#define DFLAGS_ARRAY_DISK           0x80   

#define DFLAGS_NEED_SWITCH          0x200  
#define DFLAGS_REMOVABLE_DRIVE      0x400  

#define DFLAGS_CDROM_DEVICE         0x1000 
#define DFLAGS_HARDDISK             0x2000 
#define DFLAGS_LS120                0x4000 
#define DFLAGS_BOOT_SECTOR_PROTECT  0x8000 

#define DFLAGS_48BIT_LBA         	0x8000000

#define DFLAGS_TAPE_RDP             0x20000 
#define DFLAGS_MEDIA_STATUS_ENABLED 0x40000 
#define DFLAGS_TAPE_DEVICE          0x80000 

#define DFLAGS_SET_CALL_BACK        0x100000
#define DFLAGS_SORT_UP              0x200000
#define DFLAGS_ARRAY_WAIT_EXEC      0x400000

#define DFLAGS_WIN_SHUTDOWN			0x1000000
#define DFLAGS_WIN_FLUSH   			0x2000000

/*Second device flags*/
#define DFLAGS_REDUCE_MODE	        0x00010000
#define DFLAGS_DEVICE_DISABLED	    0x00020000	
#define DFLAGS_BOOTABLE_DEVICE		0x00080000	
#define DFLAGS_BOOT_MARK			0x00100000  
#define DFLAGS_NEW_ADDED			0x40000000  

#define DFLAGS_REMAINED_MEMBER		0x80000000

/***************************************************************************
 * Description: _Channel
 */
typedef struct _Channel {
	PIDE_REGISTERS_1 BaseIoAddress1;  
	PIDE_REGISTERS_2 BaseIoAddress2;  
	PUCHAR           BMI;             
	PUCHAR           BaseBMI;         

	PUCHAR			 NextChannelBMI;  
	PUCHAR			 MiscControlAddr; 

	UCHAR            InterruptLevel;  
	UCHAR            ChannelFlags;    

	UCHAR            RetryTimes;      
	UCHAR            nSector;         
	ULONG            Lba;             

	ADDRESS          BufferPtr;       
	UINT             WordsLeft;       

	UINT             exclude_index;   

	PDevice          pWorkDev;        

	PCI1_CFG_ADDR    pci1_cfg;        
	ULONG            *Setting;        

	PDevice          pDevice[2];      
	Device           Devices[2];

	ULONG            SgPhysicalAddr;  
	PSCAT_GATH       pSgTable;        

	UCHAR            SGBuf[MAX_SG_DESCRIPTORS * sizeof(SCAT_GATH) + 16];

	PSCSI_REQUEST_BLOCK CurrentSrb;   
	SCSI_REQUEST_BLOCK  _Srb;         

	ULONG			 RetryCount;     

	struct timer_list *retry_timer;
	struct timer_list *watch_timer;
	struct timer_list *busy_timer;

	char	adjusted;    
} Channel, *PChannel;

typedef struct _Adapter {
	char *  name;
	UCHAR   chip_type;
	UCHAR   num_buses;
	UCHAR	irq;
	UCHAR	irqOwned;
	UCHAR   pci_bus;
	UCHAR	devfn;
	UCHAR	offline;
	UCHAR   reserved2;
	Channel *IDEChannel;
}
HPT_ADAPTER, *PHPT_ADAPTER;

/***************************************************************************
 * Description: Virtual Device Table
 */
typedef struct _VirtualDevice {
    UCHAR   nDisk;             
    UCHAR   BlockSizeShift;    
    UCHAR   arrayType;         
    UCHAR   pad_1;
    
    WORD    ArrayNumBlock;     
    WORD    pad_2;

    ULONG   Stamp;             
	ULONG	MirrorStamp;	

    ULONG   capacity;          

    PSCSI_REQUEST_BLOCK Srb;   

	ULONG	RaidFlags;	

	struct _Device  *pDevice[MAX_MEMBERS]; 

	UCHAR				ArrayName[32];	

	ULONG   RebuiltSector;	
	
	ULONG   LockedLba;      
	ULONG   LockedSectors;  

} VirtualDevice, *PVirtualDevice;

/*
 * RAID FLAGS declare area
 */
#define RAID_FLAGS_NEED_SYNCHRONIZE		0x00000001
#define RAID_FLAGS_NEED_REBUILD 		RAID_FLAGS_NEED_SYNCHRONIZE
#define RAID_FLAGS_BROKEN               0x00000002
#define RAID_FLAGS_BEING_BUILT			0x00000004
#define RAID_FLAGS_DISABLED				0x00000008
#define RAID_FLAGS_BOOTDISK				0x00000010
#define RAID_FLAGS_NEWLY_CREATED		0x00000020
#define RAID_FLAGS_ABORT_REBUILD		0x00000040
#define RAID_FLAGS_DUPLICATE_AND_CREATE 0x00000080

#define RAID_FLAG_SET(pArray, f) (pArray)->RaidFlags |= (f)
#define RAID_FLAG_CLEAR(pArray, f) (pArray)->RaidFlags &= ~(f)
#define RAID_FLAG_ISSET(pArray, f) (((pArray)->RaidFlags & (f)) !=0)

/*
 * The time when the error happend
 */
typedef struct _Time_Record {
   UINT    seconds:6;      /* 0 - 59 */
   UINT    minutes:6;      /* 0 - 59 */
   UINT    month:4;        /* 1 - 12 */
   UINT    hours:6;        /* 0 - 59 */
   UINT    day:5;          /* 1 - 31 */
   UINT    year:5;         /* 0=2000, 31=2031 */
} TimeRecord;

/*
 * Error Record
 */
typedef struct _ErrorLog {
   TimeRecord  Time;           /* the time when the error happend */
   UCHAR       ErrorReason;    /* Error Reason or command, see below */
   UCHAR       DiskSeq;        /* Which disk has the error */
   UCHAR       AtaStatus;      /* Ata error status (read from 1x7) */
   UCHAR       nSectors;       /* the number of sectors for I/O command */
   LONG        LBA;            /* the LBA of error location on the disk */
} ErrorLog, *PErrorLog;

#define MAX_ERROR_LOG  32      /* Maximum error log in the info block */

#define MIRROR_DISK    (MAX_MEMBERS - 1)

/* arrayType */
#define VD_RAID_0_STRIPE     0 
#define VD_RAID_1_MIRROR     1 
#define VD_RAID_01_2STRIPE   2 
#define VD_SPAN              3 
#define VD_RAID01_MIRROR     7 
#define VD_INVALID_TYPE		0xFF

extern UINT  hpt_NumAdapters;                 /* Number of Adapters */
#define NumAdapters hpt_NumAdapters
extern PHPT_ADAPTER hpt_adapters[];
extern struct Scsi_Host *hpt_vhost; /* virtual scsi host */

extern VirtualDevice VirtualDevices[MAX_V_DEVICE];
extern PVirtualDevice pLastVD;

typedef struct _logical_device {
	UCHAR   isArray;
	UCHAR   isValid;
	UCHAR	isInUse;
	UCHAR   reserved;
	PVOID   pLD;
}
LOGICAL_DEVICE;

extern LOGICAL_DEVICE hpt_logical_devices[];
extern DWORD hpt_num_logical_devices;
extern ULONG hpt_page_offset;

typedef struct _hpt_GET_INFO {
	char *buffer;
	int buflength;
	int bufoffset;
	int buffillen;
	int filpos;	
} HPT_GET_INFO;

#define MAX_DPC 16
typedef struct st_HPT_DPC {
	void (*dpc)(void *);
	void *arg;
} ST_HPT_DPC;

extern volatile int stop_thread;
extern struct semaphore * hpt_notify_wt;
extern struct semaphore * hpt_queue_ready;
extern struct task_struct *hpt_worker_thread_handle; // our kernel thread
typedef void (*HPT_DPC)(void*);

extern ST_HPT_DPC FastDpcQueue[MAX_DPC];
extern volatile int FastDpcQueue_First;
extern volatile int FastDpcQueue_Last;

/*
 * time represented in DWORD format
 */
typedef struct _TIME_RECORD {
   UINT        seconds:6;      /* 0 - 59 */
   UINT        minutes:6;      /* 0 - 59 */
   UINT        month:4;        /* 1 - 12 */
   UINT        hours:6;        /* 0 - 59 */
   UINT        day:5;          /* 1 - 31 */
   UINT        year:5;         /* 0=2000, 31=2031 */
} TIME_RECORD;

typedef DWORD DEVICEID;

/* 
 * event structure
 */
typedef struct _HPT_EVENT {
	TIME_RECORD Time;
	DEVICEID    DeviceID;
	UCHAR       EventType;
	UCHAR		reserved[3];

	UCHAR		Data[32]; /* various data depend on EventType */
} HPT_EVENT, *PHPT_EVENT;

#define MAX_EVENTS 10
extern HPT_EVENT hptdev_event_queue[];
volatile extern int event_queue_head, event_queue_tail;


/* Macros */
#define KMSG_LEADING PROC_DIR_NAME": "
#define hpt_printk(_x_) do { printk(KMSG_LEADING); printk _x_ ; } while (0)

/* DebugPrint */
#ifdef DEBUG
extern int hpt_dbg_level;
#define KdPrintE(_x_) do { if (hpt_dbg_level>0) { printk(KMSG_LEADING); printk _x_ ;} } while (0)
#define KdPrintW(_x_) do { if (hpt_dbg_level>1) { printk(KMSG_LEADING); printk _x_ ;} } while (0)
#define KdPrintI(_x_) do { if (hpt_dbg_level>2) { printk(KMSG_LEADING); printk _x_ ;} } while (0)
#else
#define KdPrintI(_x_)
#define KdPrintW(_x_)
#define KdPrintE(_x_)
#endif // DEBUG
#define KdPrint(_x_) KdPrintI(_x_)

#define MapToLogicAddr(x)            ((void *)(hpt_page_offset | (unsigned long)(x)))
#define MapToPhysAddr(x)             ((unsigned long)(x) & ~hpt_page_offset)
#define ZeroMemory(p, size)			memset(p, 0, size)

#define MAX_RETRY_COUNT 5
#define ScheduleRetryProcess(pChan) do {\
			pChan->retry_timer->expires = jiffies + 10; \
			add_timer(pChan->retry_timer);\
		} while (0)
#define CancelRetryProcess(pChan) del_timer(pChan->retry_timer)

/*
 * Busy-handle Timer, Hz=ticks/second of the clock
 */
#define SetBusyTimer(pChan) do {\
			pChan->busy_timer->expires = jiffies + (HZ>>10) + 1; \
			add_timer(pChan->busy_timer);\
		} while (0)
#define CancelBusyTimer(pChan) del_timer(pChan->busy_timer)

extern void OS_Busy_Handle(PChannel pChan, PDevice pDev);


/***************************************************************************
 * Function prototype
 */
/* array */
void Final_Array_Check(void);

/* device */
UCHAR DeviceInterrupt(PDevice pDev, UINT flags);

/* entry */
extern void hpt_handler(int irq, void *dev_id);
extern void Set_Logical_Drives(void);
extern void hpt3xx_init_post(PHPT_ADAPTER pAdap);

/* event */
void hpt_busy_timer(PChannel pChan);
void hpt_watch_timer(PChannel pChan);
extern void AtapiCallBack(IN PChannel pChan);
int event_queue_add(PHPT_EVENT pEvent);
int event_queue_remove(PHPT_EVENT pEvent);	

/* hptproc */
int hpt_set_asc_info(char *buffer, int length);
int hpt_set_info(char *buffer,int length);
void hpt_copy_mem_info(HPT_GET_INFO *pInfo, char *data, int datalen);
int hpt_copy_info(HPT_GET_INFO *pinfo, char *fmt, ...);
void get_array_name(char *arrayname, PVirtualDevice pArray);
void hpt_copy_array_info(HPT_GET_INFO *pinfo, int nld, PVirtualDevice pArray);

/* hptchip */
void SetHptChip(PChannel Primary, PUCHAR BMI);

/* hptutil */
BOOL device_rdwr(PDevice pDev, int cmd, ULONG lba, UCHAR nSector, PVOID buffer, DWORD flags);
void hpt_worker_thread(void *);
int hpt_queue_dpc(HPT_DPC dpc, void *arg);
int hpt_queue_fastdpc(HPT_DPC dpc, void *arg);
#ifdef DRIVER_REBUILD
void hpt_rebuild_mirror(PVirtualDevice pArray);
#endif
void hpt_refresh_logical_devices(void);
void hpt_set_device_offline(int id);

/* init */
BOOLEAN AtapiHwInitialize(IN PChannel pChan);

/* io */
void IdeSendCommand(IN PDevice pDev, IN PSCSI_REQUEST_BLOCK Srb);

/* ioctl */
int Kernel_DeviceIoControl(
							DWORD dwIoControlCode,       // operation control code
							PVOID lpInBuffer,            // input data buffer
							DWORD nInBufferSize,         // size of input data buffer
							PVOID lpOutBuffer,           // output data buffer
							DWORD nOutBufferSize,        // size of output data buffer
							PDWORD lpBytesReturned      // byte count
						);

/* task */
void hptretry(PChannel pChan);
void TaskQueue_FDQ(void);
BOOLEAN Driver_Busy(void);
BOOLEAN Driver_Free(void);
void Driver_Done(void);
PSCSI_REQUEST_BLOCK AllocateSrb(PChannel pChan);
void TaskQueue(void);
int CopySgl(PSCAT_GATH pSG, Scsi_Cmnd *pREQ, int isphy);
VOID TaskDone(PSCSI_REQUEST_BLOCK Srb);
BOOLEAN AtapiStartIo(PDevice pDev, PSCSI_REQUEST_BLOCK Srb);

extern void StallExec(ULONG x);
extern void SetWatchDogTimer(PChannel pChan);
extern void CancelWatchDogTimer(PChannel pChan);
extern unsigned long spin_lock_irqsave_Rhpt(int lock);
extern void spin_unlock_irqrestore_Rhpt(int lock, unsigned long flags);

#define IO_REQUEST_LOCK		0
#define HPTIO_REQUEST_LOCK	1
#define GIO_REQUEST_LOCK	2

#pragma pack()

/*
 * kernel function wrapper
 */
extern __attribute__((regparm(0))) int printk_Rhpt(const char * fmt, ...) __attribute__ ((format (printf, 1, 2)));

extern void outl_Rhpt(unsigned int value, unsigned short port);
extern unsigned int inl_Rhpt(unsigned short port);
extern void outb_Rhpt(unsigned char value, unsigned short port);
extern unsigned char inb_Rhpt(unsigned short port);
extern void outw_Rhpt(unsigned short value, unsigned short port);
extern unsigned short inw_Rhpt(unsigned short port);
extern void insw_Rhpt(unsigned short port, void * addr, unsigned long count);
extern void outsw_Rhpt(unsigned short port, const void * addr, unsigned long count);

extern unsigned long volatile hpt_get_jiffies(void);
extern int hpt_get_HZ(void);
	
extern void *kmalloc_Rhpt(unsigned int size,int priority);
extern void kfree_Rhpt(const void *objp);

extern void schedule_Rhpt(void);

#endif //_HPTGLB_H_
